#include "RoadConnector.h"
#include "../SampleMath.h"
#include "../SampleUtils.h"
#include <math.h>

Texture* RoadConnector::roadTexture = NULL;
Texture* RoadConnector::sideTexture = NULL;
float RoadConnector::EXIT_SPEED = 20;
float RoadConnector::LIFT_HEIGHT = 6;
float RoadConnector::LIFT_WIDTH = 50;

RoadConnector::RoadConnector(RoadPiece* src, RoadPiece* dst): srcRP(src), dstRP(dst){
	isValid = false;
};

float* RoadConnector::getConnectorPoints(){
	return connectorPoints;
}

void RoadConnector::generateRoadConnection(bool lift, int* points){

	// if we can't see once of the road pieces, don't draw the spline
	if (!srcRP->isRoadShowing() || !dstRP->isRoadShowing()){
		return;
	}

	if (!isValid){
		findBestConnectionPoints();
	}

	QCAR::Matrix44F srcModelView = srcRP->getModelView();
	QCAR::Matrix44F dstModelView = dstRP->getModelView();
	QCAR::Matrix44F mainModelViewInverse = SampleMath::Matrix44FInverse(srcModelView);
	QCAR::Matrix44F modelViewTranspose = SampleMath::Matrix44FTranspose(dstModelView); // let's work with row-major matrices
	QCAR::Matrix44F offsetMatrix = QCAR::Tool::multiply(mainModelViewInverse, modelViewTranspose);

	//offsetMatrix = lowPassMatrix(offsetMatrix, oldOffsetMatrix);
	//oldOffsetMatrix = offsetMatrix;
	
	// destination start point dst frame
	QCAR::Vec4F dst_pt(dstPt[0], dstPt[1], dstPt[2], 1.0f);
	// destination end point dst frame (destination direction is reversed)
	QCAR::Vec4F dst_end(dstPt[0] - dstPtDir[0], dstPt[1] - dstPtDir[1], dstPt[2] - dstPtDir[2], 1.0f);
	
	// destination start point src frame
	dst_pt = SampleMath::Vec4FTransform(dst_pt, offsetMatrix);
	// destination end point src frame
	dst_end = SampleMath::Vec4FTransform(dst_end, offsetMatrix);

	// destination direction src frame (should be length = 1 because of the logic above)
	QCAR::Vec3F dst_dir(dst_end.data[0] - dst_pt.data[0], dst_end.data[1]- dst_pt.data[1], dst_end.data[2]- dst_pt.data[2]);

	//{src_pt, end_pt, src_dir, end_dir}
	GLfloat control[12] = {srcPt[0], srcPt[1], srcPt[2],
							dst_pt.data[0], dst_pt.data[1], dst_pt.data[2],
							EXIT_SPEED*srcPtDir[0], EXIT_SPEED*srcPtDir[1], EXIT_SPEED*srcPtDir[2], 
							EXIT_SPEED*dst_dir.data[0], EXIT_SPEED*dst_dir.data[1], EXIT_SPEED*dst_dir.data[2]};

	// calculate main spline
	calculateHermitePoints(control, connectorPoints, CONNECTION_POINTS);

	if (lift && points != NULL){

		float centerX = (connectorPoints[3*points[2]] + connectorPoints[3*points[3]])/2.0f;
		float centerY = (connectorPoints[3*points[2] + 1] + connectorPoints[3*points[3] + 1])/2.0f;
		float centerZ = (connectorPoints[3*points[2] + 2] + connectorPoints[3*points[3] + 2])/2.0f;

		for (int i=1; i < RoadConnector::CONNECTION_POINTS - 1; i++){
			float distX = connectorPoints[3*i] - centerX;
			float distY = connectorPoints[3*i + 1] - centerY;
			float distZ = connectorPoints[3*i + 2] - centerZ;
			float distance = sqrt (distX*distX + distY*distY + distZ*distZ);
			// gaussian function
			float addition = LIFT_HEIGHT*exp( -(distance*distance)/LIFT_WIDTH );

			connectorPoints[3*i + 1] += addition;
		}
	}

	// calculate road geometry based on spline
	calculateRoadGeometry(srcPtDir, dst_dir, offsetMatrix);
}

void RoadConnector::liftRoad(){
	generateRoadConnection(true);
}

void RoadConnector::calculateRoadGeometry(const QCAR::Vec3F& srcDir, const QCAR::Vec3F& dstDir, QCAR::Matrix44F& offsetMat){
	// first point special case (we'll use the src point and direction for the first points)
	QCAR::Vec3F v1(srcDir.data[0], srcDir.data[1], srcDir.data[2]);
	QCAR::Vec3F up(0, 1, 0);
	v1 = SampleMath::Vec3FNormalize(SampleMath::Vec3FCross(v1, up));

	connectorUpperBounds[0][0] = connectorPoints[0] + 2.5f * v1.data[0];
	connectorUpperBounds[0][1] = connectorPoints[1] + 2.5f * v1.data[1];
	connectorUpperBounds[0][2] = connectorPoints[2] + 2.5f * v1.data[2];

	connectorLowerBounds[0][0] = connectorPoints[0] - 2.5f * v1.data[0];
	connectorLowerBounds[0][1] = connectorPoints[1] - 2.5f * v1.data[1];
	connectorLowerBounds[0][2] = connectorPoints[2] - 2.5f * v1.data[2];

	// lower point is - 1f * up;
	connectorUpperBounds[1][0] = connectorPoints[0] + 2.5f * v1.data[0];
	connectorUpperBounds[1][1] = connectorPoints[1] + 2.5f * v1.data[1] - 1.0f;
	connectorUpperBounds[1][2] = connectorPoints[2] + 2.5f * v1.data[2];

	connectorLowerBounds[1][0] = connectorPoints[0] - 2.5f * v1.data[0];
	connectorLowerBounds[1][1] = connectorPoints[1] - 2.5f * v1.data[1] - 1.0f;
	connectorLowerBounds[1][2] = connectorPoints[2] - 2.5f * v1.data[2];

	float pts[4*3];
	QCAR::Vec3F v2;
	int index;

	for (int i=0; i < CONNECTION_POINTS - 2; i++){

		// keep the old direction
		v2 = v1;

		//calculate the new direction
		v1.data[0] = connectorPoints[3*(i+2)] - connectorPoints[3*(i+1)];
		v1.data[1] = connectorPoints[3*(i+2) + 1] - connectorPoints[3*(i+1) + 1];
		v1.data[2] = connectorPoints[3*(i+2) + 2] - connectorPoints[3*(i+1) + 2];

		// find perpendicular direction
		v1 = SampleMath::Vec3FNormalize(SampleMath::Vec3FCross(v1, up));

		// find rect points
		index = 3*(i + 1);

		pts[0] = connectorPoints[index] - 2.5f * v1.data[0];
		pts[1] = connectorPoints[index + 1] - 2.5f * v1.data[1];
		pts[2] = connectorPoints[index + 2] - 2.5f * v1.data[2];

		pts[3] = connectorPoints[index] + 2.5f * v1.data[0];
		pts[4] = connectorPoints[index + 1] + 2.5f * v1.data[1];
		pts[5] = connectorPoints[index + 2] + 2.5f * v1.data[2];

		pts[6] = connectorPoints[index] - 2.5f * v2.data[0];
		pts[7] = connectorPoints[index + 1] - 2.5f * v2.data[1];
		pts[8] = connectorPoints[index + 2] - 2.5f * v2.data[2];

		pts[9] = connectorPoints[index] + 2.5f * v2.data[0];
		pts[10] = connectorPoints[index + 1] + 2.5f * v2.data[1];
		pts[11] = connectorPoints[index + 2] + 2.5f * v2.data[2];

		// store middle points of rect points as geometry
		connectorUpperBounds[0][index] = (pts[3] + pts[9])/2.0f;
		connectorUpperBounds[0][index + 1] = (pts[4] + pts[10])/2.0f;
		connectorUpperBounds[0][index + 2] = (pts[5] + pts[11])/2.0f;

		connectorLowerBounds[0][index] = (pts[6] + pts[0])/2.0f;
		connectorLowerBounds[0][index + 1] = (pts[1] + pts[7])/2.0f;
		connectorLowerBounds[0][index + 2] = (pts[2] + pts[8])/2.0f;

		connectorUpperBounds[1][index] = (pts[3] + pts[9])/2.0f;
		connectorUpperBounds[1][index + 1] = (pts[4] + pts[10])/2.0f - 1.0f;
		connectorUpperBounds[1][index + 2] = (pts[5] + pts[11])/2.0f;

		connectorLowerBounds[1][index] = (pts[6] + pts[0])/2.0f;
		connectorLowerBounds[1][index + 1] = (pts[1] + pts[7])/2.0f - 1.0f;
		connectorLowerBounds[1][index + 2] = (pts[2] + pts[8])/2.0f;
	}

	// last point special case (we'll use the dst point and direction for the last points)
	v1.data[0] = dstDir.data[0];
	v1.data[1] = dstDir.data[1];
	v1.data[2] = dstDir.data[2];
	
	QCAR::Vec4F destUp(0, 1, 0, 0);
	destUp = SampleMath::Vec4FTransform(destUp, offsetMat);
	up = SampleMath::Vec3FNormalize(QCAR::Vec3F(destUp.data[0], destUp.data[1], destUp.data[2]));
	v1 = SampleMath::Vec3FNormalize(SampleMath::Vec3FCross(v1, up));

	index = 3*(CONNECTION_POINTS - 1);

	connectorUpperBounds[0][index] = connectorPoints[index] + 2.5f * v1.data[0];
	connectorUpperBounds[0][index + 1] = connectorPoints[index + 1] + 2.5f * v1.data[1];
	connectorUpperBounds[0][index + 2] = connectorPoints[index + 2] + 2.5f * v1.data[2];

	connectorLowerBounds[0][index] = connectorPoints[index] - 2.5f * v1.data[0];
	connectorLowerBounds[0][index + 1] = connectorPoints[index + 1] - 2.5f * v1.data[1];
	connectorLowerBounds[0][index + 2] = connectorPoints[index + 2] - 2.5f * v1.data[2];

	// lower point is - 1f * dest_up;
	connectorUpperBounds[1][index] = connectorPoints[index] + 2.5f * v1.data[0] - up.data[0];
	connectorUpperBounds[1][index + 1] = connectorPoints[index + 1] + 2.5f * v1.data[1] - up.data[1];
	connectorUpperBounds[1][index + 2] = connectorPoints[index + 2] + 2.5f * v1.data[2] - up.data[2];

	connectorLowerBounds[1][index] = connectorPoints[index] - 2.5f * v1.data[0] - up.data[0];
	connectorLowerBounds[1][index + 1] = connectorPoints[index + 1] - 2.5f * v1.data[1] - up.data[1];
	connectorLowerBounds[1][index + 2] = connectorPoints[index + 2] - 2.5f * v1.data[2] - up.data[2];
}

void RoadConnector::calculateHermitePoints(GLfloat* control, GLfloat* result, int steps){
	float s,s2,s3,h1,h2,h3,h4;
	for (int t=0; t < steps; t++)
	{
		s = (float)t / (float)(steps - 1);    // scale s to go from 0 to 1
		s2 = s*s;
		s3 = s*s*s;
		h1 = 2*s3 - 3*s2 + 1;       // calculate basis function 1
		h2 = -2*s3 + 3*s2;          // calculate basis function 2
		h3 = s3 - 2*s2 + s;         // calculate basis function 3
		h4 = s3 -  s2;              // calculate basis function 4
		result[3*t] = h1*control[0] + h2*control[3] + h3*control[6] + h4*control[9];
		result[3*t + 1] = h1*control[1] + h2*control[4] + h3*control[7] + h4*control[10];
		result[3*t + 2] = h1*control[2] + h2*control[5] + h3*control[8] + h4*control[11];
	}
}

QCAR::Matrix44F RoadConnector::lowPassMatrix(const QCAR::Matrix44F& newMatrix, const QCAR::Matrix44F& oldMatrix){
	/*
	QCAR::Matrix44F r;

	for (int i=0; i<16; i++){
		r.data[i] = newMatrix.data[i]*0.5f + oldMatrix.data[i]*0.5;
	}

	return r;
	*/
	return newMatrix;
}

void RoadConnector::findBestConnectionPoints(){
	const float* srcPts = srcRP->getConPts();
	const float* dstPts = dstRP->getConPts();
	const float* srcPtsDirs = srcRP->getConPtsDirs();
	const float* dstPtsDirs = dstRP->getConPtsDirs();

	QCAR::Matrix44F srcModelView = srcRP->getModelView();
	QCAR::Matrix44F dstModelView = dstRP->getModelView();
	QCAR::Matrix44F mainModelViewInverse = SampleMath::Matrix44FInverse(srcModelView);
	QCAR::Matrix44F modelViewTranspose = SampleMath::Matrix44FTranspose(dstModelView); // let's work with row-major matrices
	QCAR::Matrix44F offsetMatrix = QCAR::Tool::multiply(mainModelViewInverse, modelViewTranspose);
	
	offsetMatrix = lowPassMatrix(offsetMatrix, oldOffsetMatrix);
	oldOffsetMatrix = offsetMatrix;

	int srcIdx = -1, dstIdx = -1;
	float minLength = FLT_MAX;
	for (int i=0; i<srcRP->getConPtsCount(); i++){
		// check for point availability
		if (srcRP->isConPtBusy(i)){
			continue;
		}
		for (int j=0; j<dstRP->getConPtsCount(); j++){
			// check for point availability
			if (dstRP->isConPtBusy(j)){
				continue;
			}
			QCAR::Vec4F dst_pt(dstPts[3*j], dstPts[3*j + 1], dstPts[3*j + 2], 1.0f);
			dst_pt = SampleMath::Vec4FTransform(dst_pt, offsetMatrix);

			float length = sqrtf((srcPts[3*i] - dst_pt.data[0])*(srcPts[3*i] - dst_pt.data[0]) 
				+ (srcPts[3*i + 1] - dst_pt.data[1])*(srcPts[3*i + 1] - dst_pt.data[1])
				+ (srcPts[3*i + 2] - dst_pt.data[2])*(srcPts[3*i + 2] - dst_pt.data[2]));

			if (length < minLength){
					minLength = length;
					srcIdx = i;
					dstIdx = j;
			}
		}
	}

	// no abavilable points
	if (srcIdx == -1 || dstIdx == -1){
		return;
	}
	
	// set src and dst points, update availability and set spline valid
	srcPt[0] = srcPts[3*srcIdx]; srcPt[1] = srcPts[3*srcIdx + 1]; srcPt[2] = srcPts[3*srcIdx + 2];
	dstPt[0] = dstPts[3*dstIdx]; dstPt[1] = dstPts[3*dstIdx + 1]; dstPt[2] = dstPts[3*dstIdx + 2];
	srcPtDir[0] = srcPtsDirs[3*srcIdx]; srcPtDir[1] = srcPtsDirs[3*srcIdx + 1]; srcPtDir[2] = srcPtsDirs[3*srcIdx + 2];
	dstPtDir[0] = dstPtsDirs[3*dstIdx]; dstPtDir[1] = dstPtsDirs[3*dstIdx + 1]; dstPtDir[2] = dstPtsDirs[3*dstIdx + 2];
	srcRP->toggleBusyConPts(srcIdx);
	dstRP->toggleBusyConPts(dstIdx);
	srcPtIdx = srcIdx;
	dstPtIdx = dstIdx;
	isValid = true;
}

RoadConnector::~RoadConnector(){
	srcRP->toggleBusyConPts(srcPtIdx);
	dstRP->toggleBusyConPts(dstPtIdx);
}

void RoadConnector::draw(QCAR::Matrix44F& projectionMatrix, const RoadGLParams& params, bool temp, bool deleting){
	
	// if we can't see once of the road pieces, don't draw the spline
	if (!srcRP->isRoadShowing() || !dstRP->isRoadShowing()){
		return;
	}

	if (!isValid){
		return;
	}

	float uv[2*CONNECTION_POINTS*2];

	for (int i=0; i< CONNECTION_POINTS; i++){
		if (i%2 == 0){
			uv[4*i] = 0.0f;
			uv[4*i+1] = 0.0f;
			uv[4*i+2] = 0.0f;
			uv[4*i+3] = 1.0f;
		}else{
			uv[4*i] = 1.0f;
			uv[4*i+1] = 0.0f;
			uv[4*i+2] = 1.0f;
			uv[4*i+3] = 1.0f;
		}
	}

	float geom[3*CONNECTION_POINTS*2];

	QCAR::Matrix44F modelViewProjection;
	SampleUtils::multiplyMatrix(&projectionMatrix.data[0], &srcRP->getModelView().data[0], &modelViewProjection.data[0]);

	if (temp && !deleting){
		glUseProgram(params.greenProgramID);
		glVertexAttribPointer(params.colorvertexHandle, 3, GL_FLOAT, GL_FALSE, 0, connectorPoints);
		glEnableVertexAttribArray(params.colorvertexHandle);	
		glUniformMatrix4fv(params.colormvpMatrixHandle, 1, GL_FALSE, (GLfloat*)&modelViewProjection.data[0]);
		glLineWidth(3.0f);
		glDrawArrays(GL_LINE_STRIP, 0, CONNECTION_POINTS);
		glDisable(GL_DEPTH_TEST);
		glDisableVertexAttribArray(params.colorvertexHandle);

		return;
	}else if (deleting){
		glUseProgram(params.redProgramID);
		glEnableVertexAttribArray(params.colorvertexHandle);	
		glUniformMatrix4fv(params.colormvpMatrixHandle, 1, GL_FALSE, (GLfloat*)&modelViewProjection.data[0]);
		glLineWidth(3.0f);
	}else{
		glUseProgram(params.objectProgramID);
		glActiveTexture(GL_TEXTURE0);
		glUniform1i(params.texSampler2DHandle, 0 /*GL_TEXTURE0*/);
		glUniformMatrix4fv(params.objectmvpMatrixHandle, 1, GL_FALSE, (GLfloat*)&modelViewProjection.data[0]);
		glVertexAttribPointer(params.normalHandle, 3, GL_FLOAT, GL_FALSE, 0, geom);
		glVertexAttribPointer(params.textureCoordHandle, 2, GL_FLOAT, GL_FALSE, 0, uv);

		glEnableVertexAttribArray(params.objectvertexHandle);
		glEnableVertexAttribArray(params.normalHandle);
		glEnableVertexAttribArray(params.textureCoordHandle);
	}

	glEnable(GL_DEPTH_TEST);


	for (int i=0; i < CONNECTION_POINTS; i++){
		geom[3*2*i] = connectorLowerBounds[0][3*i];
		geom[3*2*i + 1] = connectorLowerBounds[0][3*i + 1];
		geom[3*2*i + 2] = connectorLowerBounds[0][3*i + 2];

		geom[3*2*i + 3] = connectorUpperBounds[0][3*i];
		geom[3*2*i + 4] = connectorUpperBounds[0][3*i + 1];
		geom[3*2*i + 5] = connectorUpperBounds[0][3*i + 2];
	}

	if (deleting){
		glVertexAttribPointer(params.colorvertexHandle, 3, GL_FLOAT, GL_FALSE, 0, geom);
	}else{
		glVertexAttribPointer(params.objectvertexHandle, 3, GL_FLOAT, GL_FALSE, 0, geom);
		glBindTexture(GL_TEXTURE_2D, roadTexture->mTextureID);
	}
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 2*CONNECTION_POINTS);

	for (int i=0; i < CONNECTION_POINTS; i++){
		geom[3*2*i] = connectorLowerBounds[1][3*i];
		geom[3*2*i + 1] = connectorLowerBounds[1][3*i + 1];
		geom[3*2*i + 2] = connectorLowerBounds[1][3*i + 2];

		geom[3*2*i + 3] = connectorUpperBounds[1][3*i];
		geom[3*2*i + 4] = connectorUpperBounds[1][3*i + 1];
		geom[3*2*i + 5] = connectorUpperBounds[1][3*i + 2];
	}

	if (deleting){
		glVertexAttribPointer(params.colorvertexHandle, 3, GL_FLOAT, GL_FALSE, 0, geom);
	}else{
		glVertexAttribPointer(params.objectvertexHandle, 3, GL_FLOAT, GL_FALSE, 0, geom);
		glBindTexture(GL_TEXTURE_2D, sideTexture->mTextureID);
	}
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 2*CONNECTION_POINTS);

	for (int i=0; i < CONNECTION_POINTS; i++){
		geom[3*2*i] = connectorUpperBounds[1][3*i];
		geom[3*2*i + 1] = connectorUpperBounds[1][3*i + 1];
		geom[3*2*i + 2] = connectorUpperBounds[1][3*i + 2];

		geom[3*2*i + 3] = connectorUpperBounds[0][3*i];
		geom[3*2*i + 4] = connectorUpperBounds[0][3*i + 1];
		geom[3*2*i + 5] = connectorUpperBounds[0][3*i + 2];
	}
	if (deleting){
		glVertexAttribPointer(params.colorvertexHandle, 3, GL_FLOAT, GL_FALSE, 0, geom);
	}else{
		glVertexAttribPointer(params.objectvertexHandle, 3, GL_FLOAT, GL_FALSE, 0, geom);
		glBindTexture(GL_TEXTURE_2D, sideTexture->mTextureID);
	}
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 2*CONNECTION_POINTS);

	for (int i=0; i < CONNECTION_POINTS; i++){
		geom[3*2*i] = connectorLowerBounds[1][3*i];
		geom[3*2*i + 1] = connectorLowerBounds[1][3*i + 1];
		geom[3*2*i + 2] = connectorLowerBounds[1][3*i + 2];

		geom[3*2*i + 3] = connectorLowerBounds[0][3*i];
		geom[3*2*i + 4] = connectorLowerBounds[0][3*i + 1];
		geom[3*2*i + 5] = connectorLowerBounds[0][3*i + 2];
	}

	if (deleting){
		glVertexAttribPointer(params.colorvertexHandle, 3, GL_FLOAT, GL_FALSE, 0, geom);
	}else{
		glVertexAttribPointer(params.objectvertexHandle, 3, GL_FLOAT, GL_FALSE, 0, geom);
		glBindTexture(GL_TEXTURE_2D, sideTexture->mTextureID);
	}
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 2*CONNECTION_POINTS);

	glDisable(GL_DEPTH_TEST);
    glDisableVertexAttribArray(params.colorvertexHandle);
}

QCAR::Matrix44F& RoadConnector::getMVMatrix(){
	return srcRP->getModelView();
}

bool RoadConnector::isConnectedTo(int idx1, int idx2){
	if((srcRP->getIdx() == idx1 && dstRP->getIdx() == idx2) 
		|| (srcRP->getIdx() == idx2 && dstRP->getIdx() == idx1)){
			return true;
	}

	return false;
}

bool RoadConnector::isShowing(){
	return srcRP->isRoadShowing() && dstRP->isRoadShowing();
}
