#include "RoadManager.h"
#include <assert.h>

#include "../SampleUtils.h"
#include "../SampleMath.h"

#include "../CubeShaders.h"

#include "Objects/cow.h"
#include "Objects/straight.h"
#include "Objects/turn.h"
#include "Objects/t.h"
#include "Objects/circle.h"

#include "StraightPiece.h"
#include "TurnPiece.h"
#include "CirclePiece.h"
#include "TPiece.h"

#include "Intersection.h"

RoadManager::RoadManager(){
	selectedRP[0] = -1;
	selectedRP[1] = -1;
	selectionState = STATE_NONE_SELECTED;
}

void RoadManager::addTexture(int index, Texture* _texture){
	this->textures[index] = _texture;
}

void RoadManager::initTextures(int size){
	textures = new Texture*[size];
	textureCount = size;
}

void RoadManager::deleteTextures(){
	if (textures != 0){    
		for (int i = 0; i < textureCount; ++i){
			delete textures[i];
            textures[i] = NULL;
        }
		textureCount =0;
        delete[] textures;
        textures = NULL;
    }
}

void RoadManager::handleTouchEvent(TouchEvent* touch, QCAR::Matrix44F& inverseProjectionMatrix, float screenWidth, float screenHeight){
	touch->isTouchWaiting = false;

	// find start point and end point of screen click in 3D eye space
    QCAR::Vec4F lineStart, lineEnd;
	projectScreenPointToPlane(QCAR::Vec2F(touch->x, touch->y), QCAR::Vec3F(0, 0, 0), QCAR::Vec3F(0, 0, 1), 
								lineStart, lineEnd, inverseProjectionMatrix, screenWidth, screenHeight);

	bool found = false;

	// for each roadpiece
	for (int i=0; i<roadPieces.size();i++){
		// if it's not showing, no need to check anything
		if (!roadPieces[i]->isRoadShowing()){
			continue;
		}

		//check if the line intersects with this roadpiece
		if (roadPieces[i]->checkForIntersection(lineStart, lineEnd)){
			// RP intersecting
			handleSelectedRP(i);
			found = true;
			break;
		}
	}

	// didn't find a RP
	if (!found){
		handleSelectedRP(-1);
	}
}

void RoadManager::handleSelectedRP(int idx){
	// user clicked on empty space
	if (idx == -1){
		if (selectionState == STATE_TWO_SELECTED){
			int src = selectedRP[0];
			int dst = selectedRP[1];
			
			// cleanup selections
			roadPieces[selectedRP[0]]->setSelected(false);
			roadPieces[selectedRP[1]]->setSelected(false);
			selectedRP[0] = -1;
			selectedRP[1] = -1;
			selectionState = STATE_NONE_SELECTED;

			for(std::vector<RoadConnector*>::iterator it = roadConnectors.begin(); it != roadConnectors.end(); ++it) {
				// check if this is an existing road if so, delete it
				if ((*it)->isConnectedTo(src, dst)){
					delete (*it);
					roadConnectors.erase(it);
					tempRoadConnector = NULL;
					return;
				}
			}

			// if bot RP's had an available point, tempRoadConnector shouldn't be null
			// oterwise there's nothing to do
			if (tempRoadConnector != NULL){
				roadConnectors.push_back(tempRoadConnector);
				tempRoadConnector = NULL;
			}
		}
		return;
	}

	// user clicked on RP no. idx. handle selection logic
	switch (selectionState){
		case STATE_NONE_SELECTED:
			selectedRP[0] = idx;
			selectionState++;
			break;
		case STATE_ONE_SELECTED:
			if (selectedRP[0] == idx){
				selectedRP[0] = -1;
				selectionState--;
			}else{
				selectedRP[1] = idx;
				selectionState++;
				
				for(std::vector<RoadConnector*>::iterator it = roadConnectors.begin(); it != roadConnectors.end(); ++it) {
					// check if this is an existing road if so, delete it
					if ((*it)->isConnectedTo(selectedRP[0], selectedRP[1])){
						tempRoadConnector = *it;
						isTempForDeletion = true;
					}
				}

				if (tempRoadConnector == NULL){
					// if both RPS have available connection points create new road
					if (roadPieces[selectedRP[0]]->hasAvailableConPts() && roadPieces[selectedRP[1]]->hasAvailableConPts()){
						tempRoadConnector = new RoadConnector(roadPieces[selectedRP[0]], roadPieces[selectedRP[1]]);
					}
					isTempForDeletion = false;
				}
			}
			break;
		case STATE_TWO_SELECTED:
			if (selectedRP[0] == idx){
				selectedRP[0] = selectedRP[1];
				selectedRP[1] = -1;
				selectionState--;

				if (!isTempForDeletion){
					delete tempRoadConnector;
				}
				tempRoadConnector = NULL;

			}else if(selectedRP[1] == idx){
				selectedRP[1] = -1;
				selectionState--;

				if (!isTempForDeletion){
					delete tempRoadConnector;
				}
				tempRoadConnector = NULL;

			}else{
				// 2 RP's are already selected. do nothing
				return;
			}
			break;
		}

		roadPieces[idx]->toggleSelected();
}

void RoadManager::projectScreenPointToPlane(QCAR::Vec2F point, QCAR::Vec3F planeCenter, QCAR::Vec3F planeNormal,
											QCAR::Vec4F &lineStart, QCAR::Vec4F &lineEnd, QCAR::Matrix44F& inverseProjMatrix,
											float screenWidth, float screenHeight)
{
    // Window Coordinates to Normalized Device Coordinates
    QCAR::VideoBackgroundConfig config = QCAR::Renderer::getInstance().getVideoBackgroundConfig();
    
    float halfScreenWidth = screenWidth / 2.0f;
    float halfScreenHeight = screenHeight / 2.0f;
    
    float halfViewportWidth = config.mSize.data[0] / 2.0f;
    float halfViewportHeight = config.mSize.data[1] / 2.0f;
    
    float x = (point.data[0] - halfScreenWidth) / halfViewportWidth;
    float y = (point.data[1] - halfScreenHeight) / halfViewportHeight * -1;
    
    QCAR::Vec4F ndcNear(x, y, -1, 1);
    QCAR::Vec4F ndcFar(x, y, 1, 1);
    
    // Normalized Device Coordinates to Eye Coordinates
    QCAR::Vec4F pointOnNearPlane = SampleMath::Vec4FTransform(ndcNear, inverseProjMatrix);
    QCAR::Vec4F pointOnFarPlane = SampleMath::Vec4FTransform(ndcFar, inverseProjMatrix);
    lineStart = SampleMath::Vec4FDiv(pointOnNearPlane, pointOnNearPlane.data[3]);
    lineEnd = SampleMath::Vec4FDiv(pointOnFarPlane, pointOnFarPlane.data[3]);
}

bool linePlaneIntersection(QCAR::Vec3F lineStart, QCAR::Vec3F lineEnd, QCAR::Vec3F pointOnPlane, QCAR::Vec3F planeNormal,
							QCAR::Vec3F &intersection){
    QCAR::Vec3F lineDir = SampleMath::Vec3FSub(lineEnd, lineStart);
    lineDir = SampleMath::Vec3FNormalize(lineDir);
    
    QCAR::Vec3F planeDir = SampleMath::Vec3FSub(pointOnPlane, lineStart);
    
    float n = SampleMath::Vec3FDot(planeNormal, planeDir);
    float d = SampleMath::Vec3FDot(planeNormal, lineDir);
    
    if (fabs(d) < 0.00001) {
        // Line is parallel to plane
        return false;
    }
    
    float dist = n / d;
    
    QCAR::Vec3F offset = SampleMath::Vec3FScale(lineDir, dist);
    intersection = SampleMath::Vec3FAdd(lineStart, offset);
}



void RoadManager::drawRoads(QCAR::State state, QCAR::Matrix44F& projectionMatrix){
	for (int i=0; i<roadPieces.size(); i++){
		roadPieces[i]->setShowing(false);
	}

	QCAR::Matrix44F modelViewProjection;
	for(int tIdx = 0; tIdx < state.getNumTrackableResults(); tIdx++)
	{
		// Get the trackable:
		const QCAR::TrackableResult* trackableResult = state.getTrackableResult(tIdx);
		QCAR::Matrix44F modelViewMatrix = QCAR::Tool::convertPose2GLMatrix(trackableResult->getPose());        

		// Check the type of the trackable:
		assert(trackableResult->getType() == QCAR::TrackableResult::MARKER_RESULT);
		const QCAR::MarkerResult* markerResult = static_cast<const QCAR::MarkerResult*>(trackableResult);
		const QCAR::Marker& marker = markerResult->getTrackable();

		RoadPiece* roadPiece = roadPieces[marker.getMarkerId()];
		roadPiece->updateModelView(modelViewMatrix);
		roadPiece->setShowing(true);
		roadPiece->draw(projectionMatrix, glParams);
	}

	for (int i=0; i<roadConnectors.size(); i++){
		if (roadConnectors[i] == tempRoadConnector){
			continue;
		}
		roadConnectors[i]->generateRoadConnection();
	}

	if (tempRoadConnector != NULL){
		tempRoadConnector->generateRoadConnection();
	}

	float temp2DConnector1[2*RoadConnector::CONNECTION_POINTS];
	float temp2DConnector2[2*RoadConnector::CONNECTION_POINTS];


	if (roadConnectors.size() >=2){
	
		for (int i=0; i < roadConnectors.size() - 1; i++){
			for (int j = (i + 1); j < roadConnectors.size(); j++){
				float* firstConnector = roadConnectors[i]->getConnectorPoints();
				float* secondConnector = roadConnectors[j]->getConnectorPoints();

				QCAR::Matrix44F srcModelView = roadConnectors[i]->getMVMatrix();
				QCAR::Matrix44F dstModelView = roadConnectors[j]->getMVMatrix();

				QCAR::Matrix44F srcInverse = SampleMath::Matrix44FInverse(srcModelView);
				QCAR::Matrix44F dstTranspose = SampleMath::Matrix44FTranspose(dstModelView); 
				QCAR::Matrix44F offsetMatrix = QCAR::Tool::multiply(srcInverse, dstTranspose);

				// transform dst points to src frame
				for (int i=0;i<RoadConnector::CONNECTION_POINTS;i++){
					temp2DConnector1[2*i] = firstConnector[3*i];
					temp2DConnector1[2*i + 1] = firstConnector[3*i + 2];

					QCAR::Vec4F v2(secondConnector[3*i], secondConnector[3*i + 1], secondConnector[3*i + 2], 1.0f);
					v2 = SampleMath::Vec4FTransform(v2, offsetMatrix);
					temp2DConnector2[2*i] = v2.data[0];
					temp2DConnector2[2*i + 1] = v2.data[2];
				}
		
				Line line1(temp2DConnector1, 0, RoadConnector::CONNECTION_POINTS - 1);
				Line line2(temp2DConnector2, 0, RoadConnector::CONNECTION_POINTS - 1);

				int points[4];

				Point intersection = isCurveIntersecting(line1, line2, points);

				if (intersection.isLegit() && roadConnectors[i]->isShowing()){
					roadConnectors[j]->generateRoadConnection(true, points);
				}
			}
		}
	}

	
	for (int i=0; i<roadConnectors.size(); i++){
		if (roadConnectors[i] == tempRoadConnector){
			continue;
		}

		roadConnectors[i]->draw(projectionMatrix, glParams);
	}
	

	if (tempRoadConnector != NULL){
		tempRoadConnector->draw(projectionMatrix, glParams, true, isTempForDeletion);
	}
}

void RoadManager::initRendering(){
	LOG("Java_com_qualcomm_QCARSamples_FrameMarkers_FrameMarkersRenderer_initRendering");

	// Define clear color
	glClearColor(0.0f, 0.0f, 0.0f, QCAR::requiresAlpha() ? 0.0f : 1.0f);

	// Now generate the OpenGL texture objects and add settings
	for (int i = 0; i < textureCount; ++i)
	{
		glGenTextures(1, &(textures[i]->mTextureID));
		glBindTexture(GL_TEXTURE_2D, textures[i]->mTextureID);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, textures[i]->mWidth,
			textures[i]->mHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE,
			(GLvoid*)  textures[i]->mData);
	}

	glParams.objectProgramID		= SampleUtils::createProgramFromBuffer(objectVertexShader, objectFragmentShader);
	glParams.greenProgramID			= SampleUtils::createProgramFromBuffer(colorVertexShader, greenFragmentShader);
	glParams.redProgramID			= SampleUtils::createProgramFromBuffer(colorVertexShader, redFragmentShader);
	glParams.objectvertexHandle		= glGetAttribLocation(glParams.objectProgramID, "vertexPosition");
	glParams.normalHandle			= glGetAttribLocation(glParams.objectProgramID,	"vertexNormal");
	glParams.textureCoordHandle		= glGetAttribLocation(glParams.objectProgramID,	"vertexTexCoord");
	glParams.objectmvpMatrixHandle	= glGetUniformLocation(glParams.objectProgramID, "modelViewProjectionMatrix");
	glParams.texSampler2DHandle		= glGetUniformLocation(glParams.objectProgramID, "texSampler2D");

	glParams.colorvertexHandle		= glGetAttribLocation(glParams.greenProgramID, "vertexPosition");
	glParams.colormvpMatrixHandle	= glGetUniformLocation(glParams.greenProgramID, "modelViewProjectionMatrix");
}

void RoadManager::initRoadPieces(QCAR::MarkerTracker* markerTracker){
	//create road pieces (marker initialized inside)
	roadPieces.push_back(new StraightPiece(markerTracker, textures[0], 0));
	roadPieces.push_back(new TurnPiece(markerTracker, textures[1], 1));
	roadPieces.push_back(new TPiece(markerTracker, textures[2], 2));
	roadPieces.push_back(new CirclePiece(markerTracker, textures[3], 3));
	RoadConnector::roadTexture = textures[5];
	RoadConnector::sideTexture = textures[6];
}

RoadManager::~RoadManager(){
	for (int i=0; i<roadConnectors.size(); i++){
		delete roadConnectors[i];
	}

	for (int i=0; i<roadPieces.size(); i++){
		delete roadPieces[i];
	}
}
