﻿
#include "RoadPiece.h"

#include "../SampleUtils.h"
#include "../SampleMath.h"

RoadPiece::RoadPiece(QCAR::MarkerTracker* markerTracker, Texture* pieceTexture, int idx):pieceIdx(idx), isShowing(false), isSelected(false), texture(pieceTexture){
	char name[6];
	sprintf(name, "Road%d",idx);
	if (!markerTracker->createFrameMarker(idx, name, QCAR::Vec2F(50,50)))
    {
        LOG("Failed to create frame marker with id: %d", idx);
    }

	for (int i=0; i<MAX_CON_PTS; i++){
		busyConPts[i] = false;
	}
}

bool RoadPiece::checkForIntersection(QCAR::Vec4F &lineStart, QCAR::Vec4F &lineEnd){
    QCAR::Matrix44F inverseModelViewMatrix = SampleMath::Matrix44FInverse(modelViewMatrix);
    
    QCAR::Vec4F nearWorld = SampleMath::Vec4FTransform(lineStart, inverseModelViewMatrix);
    QCAR::Vec4F farWorld = SampleMath::Vec4FTransform(lineEnd, inverseModelViewMatrix);

	QCAR::Vec3F start, end;

    start = QCAR::Vec3F(nearWorld.data[0], nearWorld.data[1], nearWorld.data[2]);
    end = QCAR::Vec3F(farWorld.data[0], farWorld.data[1], farWorld.data[2]);

	return checkIntersectionLine(start, end);
}

bool RoadPiece::checkIntersectionLine(QCAR::Vec3F& pointA, QCAR::Vec3F& pointB)
{
    // Use the separating axis theorem to determine whether or not
    // the line intersects the object-oriented bounding box
    
    QCAR::Vec3F lineDir = SampleMath::Vec3FSub(pointB, pointA);
    
    QCAR::Vec3F xAxis(1.0f, 0.0f, 0.0f);
    QCAR::Vec3F yAxis(0.0f, 1.0f, 0.0f);
    QCAR::Vec3F zAxis(0.0f, 0.0f, 1.0f);
    
    QCAR::Vec3F crossA1 = SampleMath::Vec3FCross(xAxis, lineDir);
    QCAR::Vec3F crossA2 = SampleMath::Vec3FCross(yAxis, lineDir);
    QCAR::Vec3F crossA3 = SampleMath::Vec3FCross(zAxis, lineDir);
    
    if (isSeparatingAxisLine(xAxis, pointA, pointB)) return false;
    if (isSeparatingAxisLine(yAxis, pointA, pointB)) return false;
    if (isSeparatingAxisLine(zAxis, pointA, pointB)) return false;
    
    if (isSeparatingAxisLine(crossA1, pointA, pointB)) return false;
    if (isSeparatingAxisLine(crossA2, pointA, pointB)) return false;
    if (isSeparatingAxisLine(crossA3, pointA, pointB)) return false;
    
    return true;
}

bool RoadPiece::isSeparatingAxisLine(QCAR::Vec3F axis, QCAR::Vec3F pointA, QCAR::Vec3F pointB)
{
    // Determine whether or not the given axis separates
    // the vertices of the bounding box and the given line
    const float* BBVerticies = getBBVertices();
    float magnitude = axis.data[0] * axis.data[0] + axis.data[1] * axis.data[1] + axis.data[2] * axis.data[2];
    if (magnitude < 0.00001) return false;
    
    float minA, maxA, minB, maxB;
	
	// dot product between axis and first BB vertex
	minA = maxA = BBVerticies[0]*axis.data[0] + BBVerticies[1]*axis.data[1] + BBVerticies[2]*axis.data[2];
    
    float p;
    
    for (int i = 1; i < 8; i++) {
        // dot product between axis and i's BB vertex
		p = BBVerticies[3*i]*axis.data[0] + BBVerticies[3*i + 1]*axis.data[1] + BBVerticies[3*i + 2]*axis.data[2];
        if (p < minA) minA = p;
        if (p > maxA) maxA = p;
    }
    
    minB = maxB = SampleMath::Vec3FDot(pointA, axis);
    p = SampleMath::Vec3FDot(pointB, axis);
    if (p < minB) minB = p;
    if (p > maxB) maxB = p;
    
    if (maxA < minB) return true;
    if (minA > maxB) return true;
    
    return false;
}

void RoadPiece::draw(QCAR::Matrix44F& projectionMatrix,const RoadGLParams& params){
	
	RoadPieceGLArrays arrays;
	// this should be implemented by inheriting classes
	setGLArrays(&arrays);

	QCAR::Matrix44F modelViewProjection;
	
	SampleUtils::scalePoseMatrix(6.25, 6.25, 6.25, &modelViewMatrix.data[0]);
	SampleUtils::rotatePoseMatrix(90, 1, 0, 0, &modelViewMatrix.data[0]);
	SampleUtils::rotatePoseMatrix(180, 0, 1, 0, &modelViewMatrix.data[0]);

	SampleUtils::multiplyMatrix(&projectionMatrix.data[0], &modelViewMatrix.data[0], &modelViewProjection.data[0]);
	glUseProgram(params.objectProgramID);

	glVertexAttribPointer(params.objectvertexHandle, 3, GL_FLOAT, GL_FALSE, 0, arrays.vertices);
	glVertexAttribPointer(params.normalHandle, 3, GL_FLOAT, GL_FALSE, 0, arrays.normals);
	glVertexAttribPointer(params.textureCoordHandle, 2, GL_FLOAT, GL_FALSE, 0, arrays.texCoords);

	glEnableVertexAttribArray(params.objectvertexHandle);
	glEnableVertexAttribArray(params.normalHandle);
	glEnableVertexAttribArray(params.textureCoordHandle);
	glEnable(GL_DEPTH_TEST);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texture->mTextureID);
	glUniformMatrix4fv(params.objectmvpMatrixHandle, 1, GL_FALSE, (GLfloat*)&modelViewProjection.data[0]);
	glUniform1i(params.texSampler2DHandle, 0 /*GL_TEXTURE0*/);
	glDrawElements(GL_TRIANGLES, arrays.numIndices, GL_UNSIGNED_SHORT, arrays.indices);

	SampleUtils::checkGlError("road piece draw");

	if (isSelected){
		glUseProgram(params.greenProgramID);

		glVertexAttribPointer(params.colorvertexHandle, 3, GL_FLOAT, GL_FALSE, 0, getConPts());
		glUniformMatrix4fv(params.colormvpMatrixHandle, 1, GL_FALSE, (GLfloat*)&modelViewProjection.data[0]);
		glDrawArrays(GL_POINTS, 0, getConPtsCount());

		glDisableVertexAttribArray(params.colorvertexHandle);
		glDisableVertexAttribArray(params.colormvpMatrixHandle);
	}

    glDisable(GL_DEPTH_TEST);
    glDisableVertexAttribArray(params.objectvertexHandle);
    glDisableVertexAttribArray(params.normalHandle);
    glDisableVertexAttribArray(params.textureCoordHandle);
}

bool RoadPiece::isConPtBusy(int idx){
	return busyConPts[idx];
}

void RoadPiece::toggleBusyConPts(int idx){
	busyConPts[idx] = !busyConPts[idx];
}

void RoadPiece::setShowing(bool showing){
	isShowing = showing;
}

bool RoadPiece::isRoadShowing(){
	return isShowing;
}

bool RoadPiece::isRoadSelected(){
	return isSelected;
}

void RoadPiece::setSelected(bool selected){
	isSelected = selected;
}

void RoadPiece::updateModelView(const QCAR::Matrix44F& modelView){
	for (int i=0; i < 16; i++){
		modelViewMatrix.data[i] = modelView.data[i];
	}
}

QCAR::Matrix44F& RoadPiece::getModelView(){
	return modelViewMatrix;
}

int RoadPiece::getIdx(){
	return pieceIdx;
}

bool RoadPiece::hasAvailableConPts(){
	for(int i=0;i<getConPtsCount();i++){
		if (!isConPtBusy(i)){
			return true;
		}
	}
	return false;
}

void RoadPiece::toggleSelected(){
	isSelected = !isSelected;
}